pub fn blockDataTransfer(comptime InstrFn: type, comptime P: bool, comptime U: bool, comptime S: bool, comptime W: bool, comptime L: bool) InstrFn {
    const Arm32 = @typeInfo(@typeInfo(@typeInfo(InstrFn).Pointer.child).Fn.params[0].type.?).Pointer.child;

    return struct {
        fn inner(cpu: *Arm32, opcode: u32) void {
            if (opcode == @byteSwap(0x030020E9)) @breakpoint();

            const rn: u4 = @truncate(opcode >> 16 & 0xF);
            const rlist: u16 = @intCast(opcode & 0xFFFF);

            const reg_count: u32 = @popCount(rlist);
            const first_in_list: u4 = @truncate(@ctz(rlist)); // note that @ctz(0x0000) and @ctz(0x0001) collide

            // U determines whether the LDM/STM transfer is made upwards (U == 1)
            // or downwards (U == 0).

            const base_addr = cpu.r[rn];

            const start_addr: u32 = if (U) blk: {
                break :blk base_addr + if (P) 4 else 0;
            } else blk: {
                break :blk base_addr - (4 * reg_count) + if (!P) 4 else 0;
            };

            const new_base_addr: u32 = if (U) blk: {
                break :blk base_addr + 4 * reg_count;
            } else blk: {
                break :blk base_addr - 4 * reg_count;
            };

            var address = start_addr;

            // On Empty List:
            //
            // ARMv4 Only: R15 is Loaded/Stored
            // ARMv4/Armv5: Rn = Rn +/- 0x40

            if (rlist == 0) {
                if (Arm32.arch == .v4t) {
                    const undefined_addr: u32 = if (U) blk: {
                        break :blk base_addr + if (P) 4 else 0;
                    } else blk: {
                        break :blk base_addr - (0x40 - if (!P) 4 else 0);
                    };

                    if (L) {
                        cpu.r[15] = cpu.read(u32, undefined_addr);
                        cpu.pipe.reload(cpu);
                    } else {
                        cpu.write(u32, undefined_addr, cpu.r[15] + 4);
                    }
                }

                cpu.r[rn] = if (U) base_addr + 0x40 else base_addr - 0x40;
                return;
            }

            // FIXME: This absolutely needs revisiting :skull:
            const r15_present = rlist >> 15 & 1 == 1;

            // var first = true;
            for (first_in_list..16) |idx| {
                const i: u4 = @intCast(idx);

                if (rlist >> i & 1 == 1) {
                    transfer(cpu, r15_present, i, address);
                    address += 4;

                    // if (W and !L and first) {
                    //     cpu.r[rn] = switch (Arm32.arch) {
                    //         .v4t => new_base_addr,
                    //         .v5te => start_addr,
                    //     };

                    //     first = false;
                    // }
                }
            }

            {
                const cond = @as(u2, @intFromBool(P)) << 1 | @intFromBool(U);

                const end_address: u32 = switch (cond) {
                    0b00 => base_addr, // Decrement After
                    0b10 => base_addr - 4, // Increment Before
                    0b01 => base_addr + (4 * reg_count) - 4, // Increment After
                    0b11 => base_addr + (4 * reg_count), // Increment Before
                };

                const actual_addr = address - 4;

                if (actual_addr != end_address) {
                    cpu.panic(
                        \\P = {}, U = {}, S = {}, W = {}, L = {}
                        \\base_addr = 0x{X:0>8}
                        \\start_addr = 0x{X:0>8}
                        \\expected_end_address = 0x{X:0>8} actual_end_address = 0x{X:0>8}
                        \\rlist = 0b{b:0>16}, first_in_list = r{}
                    , .{ P, U, S, W, L, base_addr, start_addr, end_address, actual_addr, rlist, first_in_list });
                }
            }

            // What happens when W is set and Rn is in the rlist? (STM)
            //
            // Armv4: Store OLD Base if Rb is FIRST entry in Rlist, otherwise store NEW base
            // Armv5: Always store OLD Base

            if (W and !L) {
                if (rlist >> rn & 1 == 0) { // rn is not in rlist
                    cpu.r[rn] = new_base_addr;
                    return;
                }

                const mask = @as(u16, 1) << rn;
                const is_first = @popCount(rlist & (mask - 1)) == 0;

                cpu.r[rn] = switch (Arm32.arch) {
                    .v4t => if (is_first) base_addr else new_base_addr,
                    .v5te => base_addr,
                };
            }

            // What happens when W is set and Rn is in the rlist? (LDM)
            //
            // ARMv4: No writeback
            // ARMv5: writeback if Rn is "the ONLY register" or NOT the LAST register

            if (W and L) {
                if (rlist >> rn & 1 == 0) { // rn is not in rlist
                    cpu.r[rn] = new_base_addr;
                    return;
                }

                switch (Arm32.arch) {
                    .v4t => {}, // No Writeback
                    .v5te => {
                        const rn_is_last = (15 - @clz(rlist)) <= rn;

                        if (reg_count == 1 or !rn_is_last) {
                            cpu.r[rn] = new_base_addr;
                        }
                    },
                }
            }
        }

        fn transfer(cpu: *Arm32, r15_present: bool, i: u5, address: u32) void {
            if (L) {
                if (S and !r15_present) {
                    // Always Transfer User mode Registers
                    cpu.setUserModeRegister(i, cpu.read(u32, address));
                } else {
                    const value = cpu.read(u32, address);

                    cpu.r[i] = value;
                    if (i == 0xF) {
                        const mask: u32 = if (Arm32.arch == .v5te) 1 else 3;
                        cpu.r[i] &= ~mask;

                        if (Arm32.arch == .v5te) cpu.cpsr.t.write(value & 1 == 1);
                        if (S) cpu.setCpsr(cpu.spsr.raw); // FIXME: before or after the reload?

                        cpu.pipe.reload(cpu);
                    }
                }
            } else {
                if (S) {
                    // Always Transfer User mode Registers
                    // This happens regardless if r15 is in the list
                    const value = cpu.getUserModeRegister(i);
                    cpu.write(u32, address, value + if (i == 0xF) 4 else @as(u32, 0)); // PC is already 8 ahead to make 12
                } else {
                    cpu.write(u32, address, cpu.r[i] + if (i == 0xF) 4 else @as(u32, 0));
                }
            }
        }
    }.inner;
}
